import argparse
import logging
import os
import sys
from logging.handlers import RotatingFileHandler

from artvee_scraper.cli.log_arg_group import JsonLogArgGroup
from artvee_scraper.cli.file_arg_group import (JsonFileArgGroup,
                                               MultiFileArgGroup)
from artvee_scraper.writer import writer_factory

from .scraper import ArtveeScraper, CategoryType, ImageSize


def parse_cli_args() -> argparse.Namespace:
    arg_parser = argparse.ArgumentParser(
        description="Scrape artwork from https://www.artvee.com"
    )
    subparsers = arg_parser.add_subparsers()

    # Register command options & parameters
    JsonLogArgGroup(subparsers).register()
    JsonFileArgGroup(subparsers).register()
    MultiFileArgGroup(subparsers).register()

    # Check that a command has been provided
    if len(sys.argv) == 1:
        arg_parser.print_help()
        arg_parser.exit()

    return arg_parser.parse_args()


def get_logger(args: argparse.Namespace) -> logging.Logger:
    handlers = None
    if log_dir := args.log_dir:
        log_file = f"{log_dir}{os.path.sep}artvee_scraper.log"
        log_max_bytes = args.log_max_size_mb * pow(1024, 2)

        rotating_file_appender = RotatingFileHandler(
            log_file, mode='a', maxBytes=log_max_bytes, backupCount=args.log_max_backups, encoding=None, delay=0)
        handlers = [rotating_file_appender]

    logging.basicConfig(
        level=getattr(logging, args.log_level),
        format="%(asctime)s.%(msecs)03d %(levelname)s [%(threadName)s] %(module)s.%(funcName)s(%(lineno)d) | %(message)s",
        datefmt="%Y-%m-%d %H:%M:%S",
        handlers=handlers,
    )

    return logging.getLogger("artvee-scraper")


def main():
    args = parse_cli_args()
    logger = get_logger(args)

    logger.debug(f"Parsed command line arguments: {vars(args)}")
    writer = writer_factory.get_instance(args.command, args)

    # Remove any duplicate categories if specified
    categories = (
        list(dict.fromkeys(args.categories)
             ) if args.categories else list(CategoryType)
    )
    scraper = ArtveeScraper(
        writer,
        worker_threads=args.worker_threads,
        categories=sorted(categories),
        image_size=ImageSize[args.image_size],
    )

    try:
        with scraper as s:
            s.start()
    except KeyboardInterrupt:
        raise SystemExit(
            "Keyboard interrupt detected; shutting down immediately...")


main()



import argparse
from abc import ABC, abstractmethod
from pathlib import Path
from typing import Optional

from artvee_scraper.scraper import CategoryType


class ArgGroup(ABC):
    """The abstract group of command line arguments for this command.

    Attributes:
        subparsers (argparse._SubParsersAction): Action object used to create subparsers.
    """

    def __init__(self, subparsers: argparse._SubParsersAction) -> None:
        """Constructs a new `ArgGroup` instance."""
        self.subparsers = subparsers

    @abstractmethod
    def get_name(self) -> str:
        """Returns the argument group name.

        Returns:
            The argument group name.

        Raises:
            NotImplementedError: If the derived class does not override this method
        """
        raise NotImplementedError

    @abstractmethod
    def get_help(self) -> str:
        """Returns the help message for this argument group.

        Returns:
            The argument group help message.

        Raises:
            NotImplementedError: If the derived class does not override this method
        """
        raise NotImplementedError

    @abstractmethod
    def add_arguments(self, subparser: argparse.ArgumentParser) -> None:
        """Populates the subparser with command line arguments associated with this group.

        Args:
            subparser: Argument subparser to populate with command line arguments.

        Raises:
            NotImplementedError: If the derived class does not override this method.
        """
        raise NotImplementedError

    def get_description(self) -> Optional[str]:
        """Brief description of this argument group.

        Returns:
            The argument group description.
        """
        pass

    def register(self) -> argparse.ArgumentParser:
        """Creates a subparser for this command with command-line arguments defined.

        Note: A command may only be registered once; a duplicate will overwrite the previous registration.

        Returns:
            The argument subparser created for this command.
        """
        subparser = self.subparsers.add_parser(
            self.get_name(), help=self.get_help(), description=self.get_description()
        )
        subparser.set_defaults(command=self.get_name())

        # Populate subparser with required program arguments
        ArgGroup._add_program_args(subparser)
        # Populate subparser with command specific arguments
        self.add_arguments(subparser)

        return subparser

    @staticmethod
    def _add_program_args(subparser: argparse.ArgumentParser) -> None:
        """Populates the subparser with required program arguments.

        Args:
            subparser: Argument subparser to populate with command line arguments.
        """
        subparser.add_argument(
            "-t",
            "--worker-threads",
            dest="worker_threads",
            default=3,
            action=IsInRangeAction,
            metavar="[1-16]",
            help="Number of worker threads (1-16)",
            type=int,
            minInclusive=1,
            maxInclusive=16,
        )
        subparser.add_argument(
            "-l",
            "--log-level",
            dest="log_level",
            default="INFO",
            choices=["DEBUG", "INFO", "WARNING", "ERROR", "CRITICAL"],
            help="Set the application log level",
        )
        subparser.add_argument(
            "-c",
            "--category",
            dest="categories",
            action="append",
            type=CategoryType,
            choices=list(CategoryType),
            help="Category of artwork to scrape",
        )
        subparser.add_argument(
            "-s",
            "--image-size",
            dest="image_size",
            default="MAX",
            type=str,
            choices=["MAX", "STANDARD"],
            help="Image size",
        )

        log_file_group = subparser.add_argument_group(
            'optional log file arguments')
        log_file_group.add_argument(
            "--log-dir",
            dest="log_dir",
            action=IsDirAction,
            help="Log file output directory",
        )
        log_file_group.add_argument(
            "--log-max-size",
            dest="log_max_size_mb",
            default=1024,  # 1 GB
            action=IsInRangeAction,  # 1MB to 10 GB
            metavar="[1-10240]",
            help="Maximum log file size in MB (1-10,240)",
            type=int,
            minInclusive=1,
            maxInclusive=10240,
        )
        log_file_group.add_argument(
            "--log-max-backups",
            dest="log_max_backups",
            default=10,
            action=IsInRangeAction,
            metavar="[0-100]",
            help="Maximum number of log files to keep (0-100)",
            type=int,
            minInclusive=0,
            maxInclusive=100,
        )


class IsInRangeAction(argparse.Action):
    def __init__(self, minInclusive: int, maxInclusive: int, *args, **kwargs):
        self.minInclusive = minInclusive
        self.maxInclusive = maxInclusive
        super(IsInRangeAction, self).__init__(*args, **kwargs)

    def __call__(self, parser, namespace, value, option_string=None):
        if not value in range(self.minInclusive, self.maxInclusive + 1):
            parser.error(
                f"argument {option_string}: invalid choice: {value} (must be in range [{self.minInclusive}-{self.maxInclusive}])")

        setattr(namespace, self.dest, value)


class IsDirAction(argparse.Action):
    def __call__(self, parser, namespace, value, option_string=None):
        path = Path(value)

        if not path.is_dir():
            parser.error(
                f"argument {option_string}: invalid choice: A directory in the path {value} does not exist")

        setattr(namespace, self.dest, value)



import json
import logging
from dataclasses import dataclass

from artvee_scraper.artwork import Artwork

from .abstract_writer import AbstractWriter

logger = logging.getLogger("artvee-scraper")


class JsonLogWriter(AbstractWriter):
    """Class for writing an artwork to the log as a JSON object.

    Attributes:
        space_level (int, optional):
            Enable pretty-printing; number of spaces to indent. Defaults to `0`.
        sort_keys (bool, optional):
            Sort the JSON keys in alphabetical order. Defaults to `False`.
        include_image (bool, optional):
            If `True`, the Base64 binary encoded image will be included in the output. Defaults to `False`.
    """

    def __init__(
        self, space_level: int = 0, sort_keys: bool = False, include_image: bool = False
    ) -> None:
        self._indent = space_level if space_level > 1 else None
        self._sort_keys = sort_keys
        self._include_image = include_image

    def write(self, artwork: Artwork) -> bool:
        """Writes the artwork to the log as a JSON object.

        Args:
            artwork:
                Artistic work to be written to the log.

        Returns:
            `True` if the artwork was written to the log successfully; otherwise `False`.
        """
        if not self._include_image:
            artwork.image = None

        logger.info(
            json.dumps(
                artwork.to_dict(),
                indent=self._indent,
                sort_keys=self._sort_keys,
                ensure_ascii=False,
            )
        )

        return True



import argparse

from artvee_scraper.writer.writer_factory import WriterType

from .arg_group import ArgGroup, IsInRangeAction


class JsonLogArgGroup(ArgGroup):
    """The group of command line arguments associated with the `JsonLogWriter`"""

    def __init__(self, subparsers: argparse._SubParsersAction) -> None:
        super().__init__(subparsers)

    def get_name(self) -> str:
        return WriterType.JSON_LOG.writer_name

    def get_help(self) -> str:
        return WriterType.JSON_LOG.description

    def add_arguments(self, subparser: argparse.ArgumentParser) -> None:
        subparser.add_argument(
            "--space-level",
            dest="space_level",
            default=0,
            action=IsInRangeAction,
            metavar="[2-6]",
            help="Enable pretty-printing; number of spaces to indent (2-6)",
            type=int,
            minInclusive=2,
            maxInclusive=6,
        )
        subparser.add_argument(
            "--sort-keys",
            dest="sort_keys",
            action="store_true",
            help="Sort JSON keys in alphabetical order",
        )
        subparser.add_argument(
            "--include-image",
            dest="include_image",
            action="store_true",
            help="Include image bytes in the output",
        )



from typing import List, Any
import copy


class ListIterator:
    """Iterate a list in either direction.

    The iterator works on a snapshot of `elements` by using a deep copy during construction.
    Changes to `elements` will not be reflected while iterating.

    Attributes:
        elements (List[Any]):
            List of elements to iterate.
        index (int, optional)
            Set the cursor to a specified index. Defaults to 0.
    """

    def __init__(self, elements: List[Any], index: int = 0):
        self._elements = copy.deepcopy(elements)
        self._cursor = index

    def has_previous(self) -> bool:
        """Determine if more elements exist in the backward direction.

        Returns:
            `True` if more elements exist; otherwise `False`.
        """
        return self._cursor != 0

    def has_next(self) -> bool:
        """Determine if more elements exist in the forward direction.

        Returns:
            `True` if more elements exist; otherwise `False`.
        """
        return self._cursor != len(self._elements)

    def next_index(self) -> int:
        """Returns the index which would be returned by a call to `next()`.

        Returns:
            Next index, or `len(elements)` if no next element exists.
        """
        return self._cursor

    def previous_index(self) -> int:
        """Returns the index which would be returned by a call to `previous()`.

        Returns:
            Previous index, or `-1` if no previous element exists.
        """
        return self._cursor - 1

    def next(self) -> Any:
        """Returns the next element and moves the cursor forward.

        Returns:
            The next element.
        Raises:
            IndexError: If no next element exists
        """
        i = self._cursor
        if i >= len(self._elements):
            raise IndexError()

        self._cursor = i + 1
        return self._elements[i]

    def previous(self) -> Any:
        """Returns the previous element and moves the cursor backward.

        Returns:
            The previous element.
        Raises:
            IndexError: If no next element exists
        """
        i = self._cursor - 1
        if i < 0:
            raise IndexError()

        self._cursor = i
        return self._elements[i]



from abc import ABC, abstractmethod

from artvee_scraper.artwork import Artwork


class AbstractWriter(ABC):
    @abstractmethod
    def write(self, artwork: Artwork) -> bool:
        raise NotImplementedError



import base64
from dataclasses import dataclass
from typing import Optional


@dataclass
class Artwork:
    """Represents an artistic work.

    Attributes:
        url (str):
            Image download URL
        title (str):
            Name of the artwork
        category (str):
            Category the work of art is depicting (ex: abstract, still-life, animals, etc)
        artist (str, optional):
            Name of the person that created this artwork. Defaults to `Unknown Artist`
        date (str, optional):
            Year or date range the artwork was completed (ex: 2012 - 2019). Defaults to `n.d.`, no date
        origin (str, optional):
            Artist nationality and lifespan (ex: Austrian, 1834-1921). Defaults to `None`
        image (bytes, optional):
            JPG formatted image bytes. Defaults to `None`
    """

    url: str
    title: str
    category: str
    artist: str = "Unknown Artist"
    date: str = "n.d."
    origin: Optional[str] = None
    image: Optional[bytes] = None

    def get_image_encoded(self) -> Optional[str]:
        """Returns the image encoded as Base64 binary.

        Returns:
            Base64 binary encoded image.
        """
        return (
            str(base64.b64encode(self.image).decode("utf-8"))
            if self.image is not None
            else None
        )

    def to_dict(self) -> dict:
        """A dict representation of this object.

        The image will be encoded in Base64 binary.

        Returns:
            This object as a dict.
        """
        d = self.__dict__
        d["image"] = self.get_image_encoded()
        return d



import concurrent.futures
import logging
import math
import re
import traceback
from enum import Enum
from typing import List, Optional

import requests
from bs4 import BeautifulSoup, Tag

from artvee_scraper.writer.file_writer import AbstractWriter

from .artwork import Artwork

logger = logging.getLogger("artvee-scraper")


class CategoryType(Enum):
    ABSTRACT = "abstract"
    FIGURATIVE = "figurative"
    LANDSCAPE = "landscape"
    RELIGION = "religion"
    MYTHOLOGY = "mythology"
    POSTERS = "posters"
    ANIMALS = "animals"
    ILLUSTRATION = "illustration"
    STILL_LIFE = "still-life"
    BOTANICAL = "botanical"
    DRAWINGS = "drawings"
    ASIAN_ART = "asian-art"

    def __str__(self):
        return self.value

    def __lt__(self, other):
        return self.value < other.value


class ImageSize(Enum):
    MAX = "https://mdl.artvee.com/hdl/"
    STANDARD = "https://mdl.artvee.com/sdl/"


class ArtveeScraper:
    _ITEMS_PER_PAGE = 70
    _PATTERN = re.compile("^(.+) *\\((.+)\\) *$")

    def __init__(
        self,
        writer: AbstractWriter,
        worker_threads: int = 3,
        categories: List[CategoryType] = list(CategoryType),
        image_size: ImageSize = ImageSize.MAX,
    ) -> None:
        self.writer = writer
        self.workers = concurrent.futures.ThreadPoolExecutor(
            max_workers=worker_threads)
        self.categories = categories
        self.download_url_prefix = image_size.value

    def __enter__(self):
        return self

    def __exit__(self, type, value, traceback):
        self.shutdown(True)

    def start(self):
        logger.info("Starting scraper for categories %s", self.categories)

        for category in self.categories:
            page_count = ArtveeScraper._num_pages_for_category(category)

            logger.info("Category %s has %d page(s)", category, page_count)
            for page in range(1, page_count + 1):
                logger.info("Processing %s (%d/%d)",
                            category, page, page_count)
                page_url = f"https://www.artvee.com/c/{category}/page/{page}/?per_page={ArtveeScraper._ITEMS_PER_PAGE}"
                artwork_list = ArtveeScraper._scrape_artwork_data(page_url)

                results = self.workers.map(self._worker_task, artwork_list)

                # Block until all submitted tasks for a page have completed
                for r in results:
                    pass

    def shutdown(self, wait: bool) -> None:
        self.workers.shutdown(wait=wait)

    def _worker_task(self, artwork: Artwork) -> bool:
        logger.info("Processing artwork %s", artwork)

        try:
            if img_link := self._image_link_from(artwork.url):
                # Download image
                img_resp = requests.get(img_link)

                if img_resp.status_code == 200:
                    # Write the artwork to destination
                    artwork.image = img_resp.content  # raw image bytes
                    return self.writer.write(artwork)

                logger.error(
                    "Failed to download artwork from URL %s; Status Code: %d",
                    img_link,
                    img_resp.status_code,
                )
            logger.error(
                "Failed to extract image link from URL %s", artwork.url)
        except Exception:
            logging.error(
                "An error occured while processing %s; %s",
                artwork,
                traceback.format_exc(),
            )

        return False

    def _image_link_from(self, artwork_url: str) -> Optional[str]:
        logger.debug("Retrieving image download link from URL %s", artwork_url)
        download_page_resp = requests.get(artwork_url)

        if download_page_resp.status_code == 200:
            soup = BeautifulSoup(download_page_resp.content, "html.parser")
            img_links = soup.find_all(
                "a",
                {
                    "class": "prem-link gr btn dis snax-action snax-action-add-to-collection snax-action-add-to-collection-downloads"
                },
            )

            # Select the correct max/standard image size link
            for link in img_links:
                link_dest = link.get("href")
                if link_dest.startswith(self.download_url_prefix):
                    return link_dest

        logger.error(
            "Failed to retrieve image download link from URL %s; Status Code: %d",
            artwork_url,
            download_page_resp.status_code,
        )

        return None

    @staticmethod
    def _num_pages_for_category(category: CategoryType) -> int:
        logger.debug(
            "Calculating the number of pages required by category '%s'", category
        )
        url = f"https://artvee.com/c/{category}/page/1/?per_page={ArtveeScraper._ITEMS_PER_PAGE}"

        try:
            resp = requests.get(url)

            if resp.status_code == 200:
                soup = BeautifulSoup(resp.content, "html.parser")

                total_items = (
                    soup.find("p", class_="woocommerce-result-count")
                    .text.strip("items")
                    .strip()
                )

                return math.ceil(int(total_items) / ArtveeScraper._ITEMS_PER_PAGE)

            logger.error(
                "Failed to retrieve total number of items from URL %s; Status Code: %d",
                url,
                resp.status_code,
            )
        except Exception:
            logging.error(
                "An error occured while retrieving total number of items from %s; %s",
                url,
                traceback.format_exc(),
            )

        return 0

    @staticmethod
    def _scrape_artwork_data(page_url: str) -> List[Artwork]:
        scraped_artwork = []

        try:
            logger.debug("Retrieving artwork metadata URL %s", page_url)
            website_resp = requests.get(page_url)

            if website_resp.status_code == 200:
                soup = BeautifulSoup(
                    website_resp.content.decode("utf-8"), "html.parser"
                )
                all_metadata_html = soup.find_all(
                    "div", {"class": "product-element-bottom"}
                )

                for meta in all_metadata_html:
                    if artwork := ArtveeScraper._parse_metadata_html(meta):
                        scraped_artwork.append(artwork)

            else:
                logger.error(
                    "Failed to retrieve website from URL %s; Status Code: %d",
                    page_url,
                    website_resp.status_code,
                )
        except Exception:
            logging.error(
                "An error occured while processing %s; %s",
                page_url,
                traceback.format_exc(),
            )

        return scraped_artwork

    @staticmethod
    def _parse_metadata_html(metadata_html: Tag) -> Optional[Artwork]:
        try:
            img_details = metadata_html.find("h3", {"class": "product-title"})

            url = img_details.a.get("href")
            title = img_details.get_text(strip=True)
            category = metadata_html.find(
                "div", {"class": "woodmart-product-cats"}
            ).get_text(strip=True)

            artwork = Artwork(url, title, category)

            if title_matcher := ArtveeScraper._PATTERN.match(title):
                artwork.title = title_matcher.group(1).strip()
                artwork.date = title_matcher.group(2).strip()

            artist = metadata_html.find(
                "div", {"class": "woodmart-product-brands-links"}
            ).get_text(strip=True)

            if artist_matcher := ArtveeScraper._PATTERN.match(artist):
                artwork.artist = artist_matcher.group(1).strip()
                artwork.origin = artist_matcher.group(2).strip()
            else:
                artwork.artist = artist

            return artwork
        except Exception:
            logger.error(
                "Failed to parse content into a valid representation; %s",
                traceback.format_exc(),
            )

        return None



import pathlib

from setuptools import find_packages, setup

here = pathlib.Path(__file__).parent.resolve()
long_description = (here / "README.md").read_text(encoding="utf-8")

requires = ["beautifulsoup4~=4.10.0", "requests~=2.27.1", "python-slugify~=5.0.2"]

setup(
    name="artvee-scraper",
    author="Zach Duclos",
    author_email="zduclos.github@gmail.com",
    description="Fetch public domain artwork from https://www.artvee.com",
    long_description=long_description,
    long_description_content_type="text/markdown",
    url="https://github.com/zduclos/artvee-scraper",
    packages=find_packages(exclude=["tests*"]),
    include_package_data=True,
    install_requires=requires,
    license="MIT",
    scripts=[],
    zip_safe=True,
    python_requires=">= 3.8",
    classifiers=[
        "Development Status :: 5 - Production/Stable",
        "Intended Audience :: Developers",
        "Natural Language :: English",
        "Environment :: Console",
        "License :: OSI Approved :: MIT License",
        "Programming Language :: Python",
        "Programming Language :: Python :: 3",
        "Programming Language :: Python :: 3.8",
        "Programming Language :: Python :: 3.9",
        "Programming Language :: Python :: 3.10",
        "Programming Language :: Python :: 3.11",
        "Operating System :: OS Independent",
    ],
    entry_points={
        "console_scripts": [
            "artvee-scraper = artvee_scraper.runner:main",
        ]
    },
    keywords="artvee, artwork, webscraper",
    project_urls={
        "Bug Reports": "https://github.com/zduclos/artvee-scraper/issues",
        "Funding": "https://artvee.com/donate/",
        "Source": "https://github.com/zduclos/artvee-scraper",
    },
)



from __future__ import annotations

from argparse import Namespace
from enum import Enum
from typing import Callable, Optional

from .abstract_writer import AbstractWriter
from .log_writer import JsonLogWriter
from .file_writer import JsonFileWriter, MultiFileWriter


def get_instance(command_name: str, args: Namespace) -> AbstractWriter:
    """Factory for creating instances of classes derived from `AbstractWriter`.

    Args:
        command_name:
            Name of the writer to create.

    Returns:
        A newly constructed writer instance.

    Raises:
        ValueError: If the `WriterType` specified by `command_name` does not exist
    """
    if writer_type := WriterType.from_str(command_name):
        creator_fn = writer_type.creator
        return creator_fn(args)

    raise ValueError(command_name)


class WriterType(Enum):
    """Represents an artistic work.

    Attributes:
        writer_name (str):
            Command name for this writer
        description (str):
            A brief description of this writer
        creator (Callable[[Namespace], AbstractWriter]):
            Function used to create a writer instance from arguments
    """

    JSON_FILE = (
        "file-json",
        "Artwork is represented as a JSON object and written to a file",
        lambda args: JsonFileWriter(
            args.dir_path,
            space_level=args.space_level,
            sort_keys=args.sort_keys,
            overwrite_existing=args.overwrite_existing,
        ),
    )
    MULTI_FILE = (
        "file-multi",
        "Artwork image and metadata are written as separate files",
        lambda args: MultiFileWriter(
            args.metadata_dir_path,
            args.image_dir_path,
            space_level=args.space_level,
            sort_keys=args.sort_keys,
            overwrite_existing=args.overwrite_existing,
        ),
    )
    JSON_LOG = (
        "log-json",
        "Artwork is output to the log as a JSON object",
        lambda args: JsonLogWriter(
            space_level=args.space_level,
            sort_keys=args.sort_keys,
            include_image=args.include_image,
        ),
    )

    def __init__(
        self,
        writer_name: str,
        description: str,
        creator: Callable[[Namespace], AbstractWriter],
    ) -> None:
        """Constructs a new `WriterType` enum instance."""
        self.writer_name = writer_name
        self.description = description
        self.creator = creator

    @staticmethod
    def from_str(writer_name: str) -> Optional[WriterType]:
        """Convenience method to return the `WriterType` enum member associated with the `writer_name`.

        Args:
            writer_name:
                Name of the enum member to retrieve.

        Returns:
            The associated `WriterType` enum member or `None` if it does not exist.
        """
        for writer in WriterType:
            if writer.writer_name == writer_name:
                return writer

        return None



from abc import ABC, abstractmethod
from typing import List, Optional

from artvee_scraper.utils import ListIterator


class Command(ABC):
    """Interface for an executable operation that can be reverted."""

    @abstractmethod
    def execute(self) -> bool:
        """Execute this operation.

        Returns:
            `True` if the command executed successfully; otherwise `False`.
        Raises:
            NotImplementedError: If the derived class does not override this method
        """
        raise NotImplementedError

    @abstractmethod
    def revert(self) -> bool:
        """Undo this operation.

        Returns:
            `True` if the command reverted successfully; otherwise `False`.
        Raises:
            NotImplementedError: If the derived class does not override this method
        """
        raise NotImplementedError


class MacroCommand(Command):
    """Execute a sequence of subcommands.

    Attributes:
        cmds (str, List[Command]):
            Registered subcommands to be executed in order. Defaults to empty list.
        index (int, optional):
            Position of the current subcommand. Defaults to `None`
    """

    def __init__(self) -> None:
        """Constructs a new `MacroCommand` instance."""
        self._cmds: List[Command] = []
        self._iter: Optional[ListIterator] = None

    def add(self, subcommand: Command) -> None:
        """Register a subcommand.

        Args:
            subcommand:
                Subcommand to be appended to the end of the traversal list.
        """
        self._cmds.append(subcommand)

    def remove(self, subcommand: Command) -> None:
        """Deregister a subcommand.

        Args:
            subcommand:
                Subcommand to be removed from the traversal list.
        """
        self._cmds.remove(subcommand)

    def execute(self) -> bool:
        """Execute each registered command.

        When a subcommand fails during traversal, the remaining subcommands will not be executed.
        To undo changes when a failure occurs, use `revert()`.

        Returns:
            `True` if all of the subcommands executed successfully; otherwise `False`.
        """
        if self._iter is None:
            self._iter = ListIterator(self._cmds)

        while self._iter.has_next():
            cmd = self._iter.next()

            if not cmd.execute():
                self._iter.previous()
                return False

        return True

    def revert(self) -> bool:
        """Revert each registered command.

        Unexecute in reverse order, beginning with the last successful subcommand.

        Returns:
            `True` if all of the subcommands reverted successfully; otherwise `False`.
        """
        if self._iter is None:
            return True

        while self._iter.has_previous():
            cmd = self._iter.previous()

            if not cmd.revert():
                self._iter.next()
                return False

        return True



import json
import logging
import os
from typing import Optional

from artvee_scraper.artwork import Artwork
from slugify import slugify

from .abstract_writer import AbstractWriter
from .command import Command, MacroCommand

logger = logging.getLogger("artvee-scraper")


class JsonFileWriter(AbstractWriter):
    """Class for writing an artwork to the filesystem as a JSON object.

    The artwork is written to the `dir_path` directory as a JSON object with the `.json` extension.

    Attributes:
        dir_path (str):
            Path to the output directory where the JSON file will be written.
        indent (int, optional):
            Enable pretty-printing; number of spaces to indent. Defaults to `0`.
        sort_keys (bool, optional):
            Sort the JSON keys in alphabetical order. Defaults to `False`.
        open_option (str):
            File mode to open the file for write access.
    """

    def __init__(
        self,
        dir_path: str,
        space_level: int = 0,
        sort_keys: bool = False,
        overwrite_existing: bool = False,
    ) -> None:
        self._dir_path = dir_path
        self._indent = space_level if space_level > 1 else None
        self._sort_keys = sort_keys
        self._open_option = "w" if overwrite_existing else "x"

    def write(self, artwork: Artwork) -> bool:
        """Writes the artwork as a JSON object to the filesystem.

        Args:
            artwork:
                Artistic work to be written to the filesystem.

        Returns:
            `True` if the artwork was written to the filesystem successfully; otherwise `False`.
        """
        try:
            # Create a clean base filename by converting to a slug
            slug = slugify(f"{artwork.artist}-{artwork.title}")

            file_path = f"{self._dir_path}{os.path.sep}{slug}.json"

            logger.debug("Writing %s to the filesystem", file_path)

            # Write file
            with open(file_path, self._open_option) as fout:
                json.dump(
                    artwork.to_dict(),
                    fout,
                    indent=self._indent,
                    sort_keys=self._sort_keys,
                    ensure_ascii=False,
                )
                logger.debug("Wrote %s to the filesystem", file_path)

            return True
        except Exception as fee:
            logger.error(
                "Failed to write %s to the filesystem; %s", file_path, fee)

        return False


class MultiFileWriter(AbstractWriter):
    """Class for writing an artwork to the filesystem as multiple files.

    The metadata is written to the `metadata_dir_path` directory as a JSON object with the `.json` extension.
    The image is written to the `image_dir_path` directory as bytes with the `.jpg` extension.

    Attributes:
        metadata_dir_path (str):
            Path to the output directory where the JSON metadata file will be written to.
        image_dir_path (str):
            Path to the output directory where the JPG image file will be written.
        indent (int, optional):
            Enable pretty-printing; number of spaces to indent. Defaults to `0`.
        sort_keys (bool, optional):
            Sort the JSON keys in alphabetical order. Defaults to `False`.
        overwrite_existing (bool, optional)
            Open the file for write access, overwriting if it already exists. Defaults to `False`
    """

    def __init__(
        self,
        metadata_dir_path: str,
        image_dir_path: str,
        space_level: int = 0,
        sort_keys: bool = False,
        overwrite_existing: bool = False,
    ) -> None:
        self._metadata_dir_path = metadata_dir_path
        self._image_dir_path = image_dir_path
        self._indent = space_level if space_level > 1 else None
        self._sort_keys = sort_keys
        self._overwrite_existing = overwrite_existing

    def write(self, artwork: Artwork) -> bool:
        """Writes the artwork and metadata as individual files.

        An attempt to rollback changes will be made in the event that both files cannot be written to the filesystem successfully.

        Args:
            artwork:
                Artistic work to be written to the filesystem.

        Returns:
            `True` if the artwork was written to the filesystem successfully; otherwise `False`.
        """
        macro_cmd = MacroCommand()
        macro_cmd.add(
            WriteImageCommand(self._image_dir_path, artwork,
                              self._overwrite_existing)
        )
        macro_cmd.add(
            WriteMetadataCommand(
                self._metadata_dir_path,
                artwork,
                self._indent,
                self._sort_keys,
                self._overwrite_existing,
            )
        )

        if not macro_cmd.execute():
            macro_cmd.revert()
            return False

        return True


class WriteImageCommand(Command):
    """Class for writing an artwork image to the filesystem.

    The image is written to the `path` directory as bytes with the `.jpg` extension.

    Attributes:
        image (bytes):
            JPG formatted image bytes.
        path (str):
            Path to the output directory where the JPG image file will be written.
        open_option (str):
            File mode to open the file for write access.
    """

    def __init__(
        self, dir_path: str, artwork: Artwork, overwrite_existing: bool
    ) -> None:
        """Constructs a new `WriteImageCommand` instance."""
        self._image = artwork.image

        # Create a clean base filename by converting to a slug
        slug = slugify(f"{artwork.artist}-{artwork.title}")
        self._path = f"{dir_path}{os.path.sep}{slug}.jpg"

        self._open_option = "wb" if overwrite_existing else "xb"

    def execute(self) -> bool:
        """Writes the artwork image to the filesystem.

        Returns:
            `True` if the image was written to the filesystem successfully; otherwise `False`.
        """
        if self._image:
            logger.debug("Writing %s to the filesystem", self._path)

            try:
                with open(self._path, self._open_option) as fout:
                    fout.write(self._image)
                    logger.debug("Wrote %s to the filesystem", self._path)

                return True
            except Exception as e:
                logger.error(
                    "Failed to write %s to the filesystem; %s", self._path, e)

        return False

    def revert(self) -> bool:
        """Deletes the artwork image from the filesystem.

        Returns:
            `True` if the image was deleted from the filesystem successfully; otherwise `False`.
        """
        logger.debug("Deleting %s from the filesystem", self._path)

        try:
            os.remove(self._path)
            return True
        except OSError as ose:
            logger.error(
                "Failed to delete %s from the filesystem; %s", self._path, ose)

        return False


class WriteMetadataCommand(Command):
    """Class for writing artwork metadata to the filesystem.

    The metadata is written to the `path` directory as bytes with the `.json` extension.

    Attributes:
        artwork (Artwork):
            Metadata describing the artistic work.
        path (str):
            Path to the output directory where the JSON metadata will be written.
        indent (int):
            Enable pretty-printing; number of spaces to indent.
        sort_keys (bool):
            Sort the JSON keys in alphabetical order.
        open_option (str):
            File mode to open the file for write access.
    """

    def __init__(
        self,
        dir_path: str,
        artwork: Artwork,
        indent: Optional[int],
        sort_keys: bool,
        overwrite_existing: bool,
    ) -> None:
        """Constructs a new `WriteMetadataCommand` instance."""
        self._artwork = artwork

        # Create a clean base filename by converting to a slug
        slug = slugify(f"{artwork.artist}-{artwork.title}")
        self._path = f"{dir_path}{os.path.sep}{slug}.json"

        self._indent = indent
        self._sort_keys = sort_keys

        self._open_option = "w" if overwrite_existing else "x"

    def execute(self) -> bool:
        """Writes this artwork metadata to the filesystem.

        Returns:
            `True` if the metadata was written to the filesystem successfully; otherwise `False`.
        """
        logger.debug("Writing %s to the filesystem", self._path)

        try:
            with open(self._path, self._open_option) as fout:
                artwork_dict = vars(self._artwork)
                artwork_dict.pop("image")

                json.dump(
                    artwork_dict,
                    fout,
                    indent=self._indent,
                    sort_keys=self._sort_keys,
                    ensure_ascii=False,
                )
                logger.debug("Wrote %s to the filesystem", self._path)

            return True
        except Exception as e:
            logger.error(
                "Failed to write %s to the filesystem; %s", self._path, e)

        return False

    def revert(self) -> bool:
        """Deletes this artwork metadata from the filesystem.

        Returns:
            `True` if the metadata was deleted from the filesystem successfully; otherwise `False`.
        """
        logger.debug("Deleting %s from the filesystem", self._path)

        try:
            os.remove(self._path)
            return True
        except OSError as ose:
            logger.error(
                "Failed to delete %s from the filesystem; %s", self._path, ose)

        return False
